@@ -1,14 +1,16 @@
+# TODO: token 相关的初始化可以通过 default Callable 来实现
 import hashlib
 import os
 import time
 from datetime import datetime
 from pathlib import Path
 from typing import BinaryIO, List, Optional, Union
+from urllib.parse import urlparse
 
 import boto3
 import requests
 
-from .dataset import Dataset, DatasetFile, UploadInfo
+from .dataset import Dataset, DatasetFile, UploadInfo, DatasetList
 import random
 import string
 
@@ -26,6 +28,19 @@ def parse_datetime(date_string: str) -> datetime:
     return datetime.strptime(date_string, "%Y-%m-%dT%H:%M:%S.%fZ")
 
 
+def convert_to_dataset_file(files) -> List[DatasetFile]:
+    dataset_files = [
+        DatasetFile(
+            file.get("_id"),
+            file.get("Token"),
+            file.get("Size"),
+            "" if file.get("SubPath") is None else file.get("SubPath"),
+        )
+        for file in files
+    ]
+    return dataset_files
+
+
 def get_dataset(dataset_id: str, token: Optional[str] = None) -> Optional[Dataset]:
     """
     Fetches dataset details from Heywhale.
@@ -50,16 +65,20 @@ def get_dataset(dataset_id: str, token: Optional[str] = None) -> Optional[Datase
 
     response = requests.get(url, headers=headers, timeout=10)
     if response.status_code == 200:
-        dataset_files = [
-            DatasetFile(file.get("_id"), file.get("Token"), file.get("Size"))
-            for file in response.json().get("Files")
-        ]
+        # print("get datasets ", response.json())
+        files = convert_to_dataset_file(response.json().get("Files"))
+        files_with_subpath = convert_to_dataset_file(
+            response.json().get("FilesStructure")
+        )
+        # print("files_with_subpath", files_with_subpath)
+        files.extend(files_with_subpath)
+        # print("return dataset files",files)
         return Dataset(
             _id=response.json().get("_id"),
             title=response.json().get("Title"),
             short_description=response.json().get("ShortDescription"),
             folder_name=response.json().get("FolderName"),
-            files=dataset_files,
+            files=files,
             created_at=parse_datetime(response.json().get("CreateDate")),
             updated_at=parse_datetime(response.json().get("UpdateDate")),
         )
@@ -85,11 +104,20 @@ def _update_dataset(id: str, files: List[DatasetFile], token: Optional[str] = No
             )
     url = f"{HEYWHALE_SITE}/api/datasets/{id}/files"
     headers = {"x-kesci-token": token, "x-kesci-resource": id}
-
+    data = {
+        "Files": [file.key for file in files],
+        "FilesStructure": [
+            {"Token": file.key, "SubPath": file.sub_path} for file in files
+        ],
+    }
+    print("data", data)
     response = requests.put(
         url,
-        {
+        json={
             "Files": [file.key for file in files],
+            "FilesStructure": [
+                {"Token": file.key, "SubPath": file.sub_path} for file in files
+            ],
         },
         headers=headers,
         timeout=10,
@@ -145,9 +173,51 @@ def generate_timestamped_string(revision: int) -> str:
     return result
 
 
+def _upload_file(
+    path_or_fileobj: Union[str, Path, bytes, BinaryIO],
+    path_in_dataset: str | Path,
+    upload_info: UploadInfo,
+)->str:
+    """
+    Uploads a file to a dataset.
+
+    :param path_or_fileobj: The path or file object of the file to upload.
+    :param path_in_dataset: The path to save the file in the dataset.
+    :param id: The ID of the dataset.
+    :param overwrite: Whether to overwrite an existing file with the same path in the dataset.
+    :param token: The token for authentication. If not provided, the function will use the 'MW_TOKEN' environment variable.
+    :raises ValueError: If no token is provided and the 'MW_TOKEN' environment variable is not set.
+    """
+
+    session = boto3.Session(
+        aws_access_key_id=upload_info.ak,
+        aws_secret_access_key=upload_info.sk,
+        aws_session_token=upload_info.token,
+        region_name=upload_info.region,
+    )
+    s3 = session.client("s3", endpoint_url=upload_info.endpoint)
+
+    bucket_name = upload_info.bucket
+
+    object_key = os.path.join(
+        upload_info.prefix_to_save, generate_timestamped_string(1)
+    )
+    if path_in_dataset != "":
+        object_key = os.path.join(object_key, path_in_dataset)
+    # print(f"upload object key {object_key}")
+    try:
+        if isinstance(path_or_fileobj, (str, Path)):
+            with open(path_or_fileobj, "rb") as file:
+                s3.put_object(Bucket=bucket_name, Key=object_key, Body=file)
+        else:
+            s3.put_object(Bucket=bucket_name, Key=object_key, Body=path_or_fileobj)
+    except Exception as e:
+        print(f"Error putting object '{object_key}' from bucket '{bucket_name}': {e}")
+    return object_key
+
 def upload_file(
     path_or_fileobj: Union[str, Path, bytes, BinaryIO],
-    path_in_dataset: str,
+    path_in_dataset: str | Path,
     id: str | Dataset,
     overwrite: bool = False,
     token: Optional[str] = None,
@@ -171,41 +241,29 @@ def upload_file(
     if isinstance(id, Dataset):
         id = id._id
     upload_info = _get_update_token(token)
-    session = boto3.Session(
-        aws_access_key_id=upload_info.ak,
-        aws_secret_access_key=upload_info.sk,
-        aws_session_token=upload_info.token,
-        region_name=upload_info.region,
-    )
-    s3 = session.client("s3", endpoint_url=upload_info.endpoint)
-
-    bucket_name = upload_info.bucket
-    prefix = upload_info.prefix_to_save
-
-    object_key = os.path.join(
-        upload_info.prefix_to_save, generate_timestamped_string(1)
-    )
-    if path_in_dataset != "":
-        object_key = os.path.join(object_key, path_in_dataset)
-    # print(f"upload object key {object_key}")
-    try:
-        if isinstance(path_or_fileobj, (str, Path)):
-            with open(path_or_fileobj, "rb") as file:
-                s3.put_object(Bucket=bucket_name, Key=object_key, Body=file)
-        else:
-            s3.put_object(Bucket=bucket_name, Key=object_key, Body=path_or_fileobj)
-    except Exception as e:
-        print(f"Error putting object '{object_key}' from bucket '{bucket_name}': {e}")
+    if upload_info is None:
+        raise ValueError("Failed to get upload token.")
+    object_key = _upload_file(path_or_fileobj, path_in_dataset, upload_info)
     new_dataset_files: List[DatasetFile] = []
     dataset = get_dataset(id, token)
     if dataset is not None:
         for file in dataset.files:
-            if _get_file_in_dataset(prefix, file.key) != path_in_dataset:
+            if _get_file_in_dataset(upload_info.prefix_to_save, file.key) != path_in_dataset:
                 new_dataset_files.append(file)
             elif not overwrite:
                 # print("file exists, skip uploading")
                 return
-        new_dataset_files.append(DatasetFile("", object_key, 0))
+        pp = str(Path(path_in_dataset).parent)
+        # 强加一个 /，按照示例格式来，怕没加没处理好。
+        if not pp.endswith("/"):
+            pp = pp + "/"
+        # 开头的也去掉，避免绝对路径的格式。
+        if pp.startswith("/"):
+            pp = pp[1:]
+        # 本地目录直接忽略
+        if pp.startswith("./"):
+            pp = pp[2:]
+        new_dataset_files.append(DatasetFile("", object_key, 0, pp))
         _update_dataset(id, new_dataset_files, token)
 
 
@@ -224,16 +282,39 @@ def upload_folder(
             )
     if isinstance(id, Dataset):
         id = id._id
+    upload_info = _get_update_token(token)
+    if upload_info is None:
+        raise ValueError("Failed to get upload token.")
+    new_dataset_files: List[DatasetFile] = []
+    dataset = get_dataset(id, token)
+    if dataset is None:
+        raise ValueError(f"Dataset '{id}' not found.")
     # Walk through the directory
     for root, _dirs, files in os.walk(folder_path):
         for file in files:
             file_path = os.path.join(root, file)
             rel_path = os.path.relpath(os.path.join(root, file), folder_path)
             # print(f"root {root} rel_path {rel_path} upload file {file_path} path in {os.path.join(path_in_dataset,rel_path)} id {id}")
-            upload_file(
-                file_path, os.path.join(path_in_dataset, rel_path), id, overwrite, token
-            )
-
+            object_key = _upload_file(
+                file_path, os.path.join(path_in_dataset, rel_path), upload_info)
+            for dfile in dataset.files:
+                if _get_file_in_dataset(upload_info.prefix_to_save, dfile.key) != path_in_dataset:
+                    new_dataset_files.append(dfile)
+                elif not overwrite:
+                    # print("file exists, skip uploading")
+                    break
+                pp = str(Path(path_in_dataset).parent)
+                # 强加一个 /，按照示例格式来，怕没加没处理好。
+                if not pp.endswith("/"):
+                    pp = pp + "/"
+                # 开头的也去掉，避免绝对路径的格式。
+                if pp.startswith("/"):
+                    pp = pp[1:]
+                # 本地目录直接忽略
+                if pp.startswith("./"):
+                    pp = pp[2:]
+                new_dataset_files.append(DatasetFile("", object_key, 0, pp))
+    _update_dataset(id, new_dataset_files, token)
 
 def _get_file_in_dataset(prefix: str, object_key: str) -> str:
     """
@@ -350,11 +431,14 @@ def download_dir(
         files = response.json().get("files")
         for file in files:
             download_url = file.get("Url")
-            token = str(file.get("Token"))
-            parts = token.split("/")
+            # Parse the URL to get components
+            parsed_url = urlparse(download_url)
+            # Extract the path component
+            # 本来返回值应该有个Token和parsed_url.path等价的但是有时候是None
+            parts = parsed_url.path.split("/")
             # Remove the first 3 parts and join the remaining parts back into a string
             remaining_parts = parts[
-                3:
+                4:
             ]  # This will slice the list to skip the first 3 elements
             # Join the remaining parts with "/" to form the final string
             filepath_in_dataset = "/".join(remaining_parts)
@@ -488,7 +572,7 @@ def create_dataset(
         "FolderName": folder_name,
         "EnableDownload": enable_download,
         "Type": 0,
-        "Files": [object_key],
+        "Files": [object_key],  # 没用FileStructure，因为本来也是个假文件
     }
 
     response = requests.post(
@@ -500,15 +584,73 @@ def create_dataset(
 
     if response.status_code == 200:
         document = response.json().get("document")
+        files = convert_to_dataset_file(document.get("Files"))
         return Dataset(
             _id=document.get("_id"),
             title=document.get("Title"),
             short_description=document.get("ShortDescription"),
             folder_name=document.get("FolderName"),
-            files=document.get("Files"),  # empty actually
+            files=files,  # empty actually
             created_at=parse_datetime(document.get("CreateDate")),
             updated_at=parse_datetime(document.get("UpdateDate")),
         )
     else:
         response.raise_for_status()
         return None
+
+
+def list_datasets(
+    title: str = "",
+    limit: int = 10,
+    start_date: time = None,
+    end_date: time = None,
+    token: Optional[str] = None,
+):
+    if token is None:
+        token = os.getenv("MW_TOKEN")
+        if not token:
+            raise ValueError(
+                "No token provided and 'MW_TOKEN' environment variable is not set."
+            )
+    url = f"{HEYWHALE_SITE}/api/datasets"
+    headers = {
+        "x-kesci-token": token,
+    }
+    response = requests.get(
+        url,
+        headers=headers,
+        params={
+            "perPage": limit,
+            "Title": title,
+            "startDate": start_date,
+            "endDate": end_date,
+        },
+        timeout=10,
+    )
+    if response.status_code == 200:
+        document = response.json()
+        datasets: List[Dataset] = []
+        for d in document.get("data"):
+            files = convert_to_dataset_file(d.get("Files"))
+            files_with_subpath = convert_to_dataset_file(d.get("FilesStructure"))
+            files.extend(files_with_subpath)
+            datasets.append(
+                Dataset(
+                    _id=d.get("_id"),
+                    title=d.get("Title"),
+                    short_description=d.get("ShortDescription"),
+                    folder_name=d.get("FolderName"),
+                    files=files,
+                    created_at=parse_datetime(d.get("CreateDate")),
+                    updated_at=parse_datetime(d.get("UpdateDate")),
+                )
+            )
+        return DatasetList(
+            datasets=datasets,
+            total=document.get("totalNum"),
+            page=document.get("page"),
+            limit=document.get("perPage"),
+        )
+    else:
+        response.raise_for_status()
+        return None
